"""Console script for pyvue."""
import os, sys, zipfile, shutil, requests
import socket, errno
import sqlite3

from .static import project_template, app_template
from .logger import Logger
from .server import Server

def main(args):
    config_file = os.path.join(os.getcwd(), ".config")
    """Console script for pyvue."""
    if args["job"] == "create-project":
        Logger.info("Creating pyvuejs project...")
        
        app_root = os.path.join(os.getcwd(), args["name"])
        if os.path.exists(app_root):
            raise RuntimeError("Project {} already exists!".format(args["name"]))
        else:
            os.mkdir(app_root)

        Logger.info("Extracting template files...")
        zp = zipfile.ZipFile(project_template)
        zp.extractall(app_root)
        zp.close()

        Logger.info("Project \"{}\" is ready!\n".format(args["name"]))
    elif args["job"] == "create-app":
        Logger.info("Creating pyvuejs app...")

        app_dir = os.path.join(os.getcwd(), args["name"])
        if os.path.exists(app_dir):
            raise RuntimeError("App \"{}\" already exists!".format(args["name"]))
        else:
            os.mkdir(app_dir)

        Logger.info("Extracting template files...")
        zp = zipfile.ZipFile(app_template)
        zp.extractall(app_dir)
        zp.close()

        Logger.info("App \"{}\" is ready!\n".format(args["name"]))
    elif args["job"] == "remove-app":
        Logger.info("Removing app \"{}\"...".format(args["name"]))

        if args["name"] == "main":
            raise RuntimeError("App \"main\" cannot be removed!")

        app_dir = os.path.join(os.getcwd(), args["name"])
        if os.path.exists(app_dir):
            shutil.rmtree(os.path.join(os.getcwd(), args["name"]))
        else:
            raise RuntimeError("App \"{}\" not exists!".format(args["name"]))

        Logger.info("App \"{}\" removed!".format(args["name"]))

    elif args["job"] == "start":
        con = sqlite3.connect(config_file, check_same_thread = False)
        cursor = con.cursor()

        if cursor.execute("select count(*) from `state`;").fetchone()[0] > 0:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                sock.bind(("127.0.0.1", int(args["port"])))
                cursor.execute("delete from `state`;")
                con.commit()
            except:
                raise RuntimeError("Server is already started!")

        cursor.execute("insert into `state` values ('{0}', {1});".format(
            args["host"], args["port"]
        ))
        con.commit()

        cursor.close()
        con.close()

        if args["mode"] == "server":
            Server(os.getcwd(), True).start(args["host"], int(args["port"]))
        elif args["mode"] == "standalone":
            Server(os.getcwd(), False).start_standalone(args["host"], int(args["port"]))
    elif args["job"] == "stop":
        con = sqlite3.connect(config_file, check_same_thread = False)
        cursor = con.cursor()

        if cursor.execute("select count(*) from `state`;").fetchone()[0] == 0:
            raise RuntimeError("Server not started!")
        
        Logger.info("Server is shutting down...")

        port = cursor.execute("select `port` from `state`;").fetchone()[0]
        requests.post("http://127.0.0.1:{}/stop".format(port))

        cursor.close()
        con.close()

if __name__ == "__main__":
    sys.exit(main(sys.argv))  # pragma: no cover



# -*- coding: utf-8 -*-
import types
from typing import Union
from inspect import signature
from threading import Thread

from .webview import Webwindow, WebDialog

class Model():
    __class_type__ = "model"
    app_view:Union[Webwindow, WebDialog] = None

    def __init__(self):
        self.__call_info = {}

        self.__variables, self.__events, self.__methods, self.__sessions = {}, {}, {}, {}
        for mname in dir(self):
            if not mname.startswith("__") and not mname.startswith("_Model__"):
                may_var = eval("self.{}".format(mname))
                if "__bind_info__" in dir(may_var):
                    bind_info = may_var.__bind_info__
                    
                    if bind_info["type"] == "variable":
                        self.__variables[mname] = may_var.value
                    elif bind_info["type"] == "session":
                        self.__sessions[mname] = may_var.value
                    elif bind_info["type"] == "method":
                        self.__methods[mname] = may_var
                    elif bind_info["type"] == "event":
                        self.__events[bind_info["name"]] = may_var

    def __set_variables(self):
        variable_names = list(self.__variables.keys())

        for v_name in variable_names:
            self.__variables[v_name] = eval("self.{}".format(v_name))

            if "__bind_info__" in dir(self.__variables[v_name]):
                self.__variables[v_name] = self.__variables[v_name].value

    @property
    def variables(self) -> dict:
        return self.__variables

    @property
    def sessions(self) -> dict:
        return self.__sessions

    @property
    def methods(self) -> dict:
        return self.__methods

    @property
    def events(self) -> dict:
        return self.__events

    def __call_function(self):
        if self.__call_info["call_type"] == "event":
            call = self.events[self.__call_info["call_name"]]
        else:
            call = self.methods[self.__call_info["call_name"]]

        call_params = signature(call).parameters
        if "session" in call_params.keys():
            call(self.__call_info["session"])
        else:
            call()

        self.__set_variables()

    def call_event(self, event_name:str, session = None):
        self.__call_info["call_type"] = "event"
        self.__call_info["call_name"] = event_name
        self.__call_info["session"] = session

        thread = Thread(target = self.__call_function)
        thread.start()

    def call_method(self, method_name:str, session = None):
        self.__call_info["call_type"] = "method"
        self.__call_info["call_name"] = method_name
        self.__call_info["session"] = session

        thread = Thread(target = self.__call_function)
        thread.start()

class View():
    def __init__(self, prefix:str, view_name:str, template:str, resources:str, style:str):
        self.__prefix = prefix
        self.__name = view_name
        self.__template = template
        self.__resources = resources
        self.__style = style

    @property
    def prefix(self) -> str:
        return self.__prefix

    @property
    def name(self) -> str:
        return self.__name

    def render(self, view_id:str, models:list) -> str:
        from .static import baseView

        return baseView.replace(
            "{$viewResource}", self.__resources
        ).replace(
            "{$viewStyle}", self.__style
        ).replace(
            "{$viewTemplate}", self.__template
        ).replace(
            "{$viewId}", view_id
        ).replace(
            "{$viewName}", self.name
        ).replace(
            "{$viewModels}", str(models)
        )



# -*- coding: utf-8 -*-
import os, sys, types
from glob import glob
from flask import Blueprint
from .logger import Logger
from .models import View, Model
import importlib


class Interpreter():
    def __init__(self, app_root:str, server, logging:bool = True):
        self.__app_root = app_root
        self.__server = server
        self.__logging = logging

    def interprete_view(self, view_file:str) -> dict:
        view_parsed = {}
        if self.__logging:
            Logger.info("Interpreting pvue file...")

        with open(view_file, "r", encoding = "utf-8") as vfr:
            vft = vfr.read()

        if vft.startswith("!prefix"):
            view_parsed["prefix"] = vft.splitlines()[0][7:].strip()
        else:
            view_parsed["prefix"] = "view"
        
        for block_name in ["resources", "style", "template"]:
            start_tag, end_tag = "<{}>".format(block_name), "</{}>".format(block_name)
            if start_tag in vft and end_tag in vft:
                view_parsed[block_name] = vft.split(start_tag)[1].split(end_tag)[0]
            else:
                view_parsed[block_name] = ""

        if self.__logging:
            Logger.info("finished")

        return view_parsed

    def load_models(self, app_name:str) -> dict:
        models_loaded = {}
        if self.__logging:
            Logger.info("Interpreting models...")

        models = importlib.import_module(".models", app_name)
        for may_model_name in [mname for mname in dir(models) if not mname == "Model" and not mname.startswith("_")]:
            # try:
            # may_model = eval("models.{}".format(may_model_name))
            may_model = getattr(models, may_model_name)
            if "__class_type__" in dir(may_model):
                models_loaded[may_model_name] = may_model()
                if self.__logging:
                    Logger.info("Model {} loaded".format(may_model_name))
            # except:
            #     pass

        if self.__logging:
            Logger.info("finished\n")

        return models_loaded

    def link_static(self):
        if os.path.exists(os.path.join(self.__app_root, "static")):
            if self.__logging:
                Logger.info("Linking static to server...")

            self.__server.register_blueprint(
                Blueprint(
                    os.path.basename(self.__app_root), "pyvue",
                    static_url_path = "/app",
                    static_folder = os.path.join(self.__app_root, "static")
                )
            )

            if self.__logging:
                Logger.info("finished\n")
        else:
            if self.__logging:
                Logger.warn("Static files are missing!\n")

    def load_app(self, app_name:str) -> dict:
        if self.__logging:
            Logger.info("Interpreting app \"{}\"...".format(app_name))

        view_parsed = self.interprete_view(os.path.join(self.__app_root, app_name, "view.html"))
        models_loaded = self.load_models(app_name)

        return {
            "view": View(
                view_parsed["prefix"], app_name,
                view_parsed["template"],
                view_parsed["resources"], view_parsed["style"]
            ),
            "models": models_loaded
        }

    def interprete(self) -> dict:
        interpreted = {}

        if self.__logging:
            Logger.info("Interpreting project \"{}\"...\n".format(os.path.basename(self.__app_root)))

        sys.path.append(self.__app_root)
        may_apps = list(os.walk(self.__app_root))[0][1]
        if not "main" in may_apps:
            raise RuntimeError("App \"main\" not exists!")

        for app_name in may_apps:
            try:
                exec("from {} import __package_type__".format(app_name))
                interpreted[app_name] = None
            except:
                pass

        for app_name in interpreted.keys():
            interpreted[app_name] = self.load_app(app_name)

        self.link_static()

        if self.__logging:
            Logger.info("Project is ready!\n")

        return interpreted



# -*- coding: utf-8 -*-

# def model(view_name:str):
#     def decorator(cls):
#         return cls

#     return decorator

class model_variable():
    __bind_info__ = { "type": "variable" }
    def __init__(self, raw_var):
        self.__raw_var = raw_var

    @property
    def value(self):
        return self.__raw_var

class session_variable():
    __bind_info__ = { "type": "session" }
    def __init__(self, raw_var):
        self.__raw_var = raw_var

    @property
    def value(self):
        return self.__raw_var

def method(func):
    def decorator(func):
        func.__bind_info__ = { "type": "method" }
        return func

    return decorator(func)

def event(event_name:str):
    def decorator(func):
        func.__bind_info__ = { "type": "event", "name": event_name }
        return func

    return decorator



import sys, argparse
from .cli import main

args = argparse.ArgumentParser(description = "manage script of pyvuejs")
args.add_argument("job", help = "job[\"create-project\", \"create-app\", \"start\", \"stop\"]")
args.add_argument("--name", help = "name for create")
args.add_argument("--host", default = "0.0.0.0", help = "host for server")
args.add_argument("--port", default = 8080, help = "port for server")
args.add_argument("--mode", default = "server", help = "mode for server[\"server\", \"standalone\"]")

if __name__ == "__main__":
    try:
        sys.exit(main(vars(args.parse_args())))
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-
import sys
from PySide2.QtWidgets import QApplication, QMainWindow, QDialog, QGridLayout
from PySide2.QtWebEngineWidgets import QWebEngineView, QWebEnginePage
from PySide2.QtCore import QUrl, Signal
from PySide2.QtGui import QIcon


__qtApp = QApplication(sys.argv)
__main_window = None
class WebDialog(QDialog):
    def __init__(self, title:str, url:str, x:int = None, y:int = None, width:int = None, height:int = None, parent = None):
        super().__init__(parent)

        self.__central_layout = QGridLayout(self)
        self.__central_layout.setContentsMargins(0, 0, 0, 0)

        self.__webview = QWebEngineView(self)
        self.__webview.setPage(QWebEnginePage(parent = self.__webview))
        self.__webview.page().loadFinished.connect(self.__onload)
        self.__central_layout.addWidget(self.__webview, 0, 0, 1, 1)

        self.setWindowTitle(title)
        self.hide()
        
        if not x == None and not y == None:
            self.move(x, y)
        
        self.resize(
            950 if width == None else width,
            650 if height == None else height
        )

        self.__webview.setUrl(QUrl(url))

    def __onload(self, ev):
        self.open()

class Webwindow(QMainWindow):
    create_window = Signal(object, str, str, int, int, int, int)
    create_dialog = Signal(object, str, str, int, int, int, int, QMainWindow)

    def __init__(self, title:str, url:str, x:int = None, y:int = None, width:int = None, height:int = None):
        super().__init__(None)
        self.create_window.connect(self.on_create_window)
        self.create_dialog.connect(self.on_create_dialog)

        self.__subwindows = []

        self.__webview = QWebEngineView(self)
        self.__webview.setPage(QWebEnginePage(parent = self.__webview))
        self.__webview.page().loadFinished.connect(self.__onload)
        self.setCentralWidget(self.__webview)

        self.setWindowTitle(title)
        self.hide()
        
        if not x == None and not y == None:
            self.move(x, y)
        
        self.resize(
            950 if width == None else width,
            650 if height == None else height
        )

        self.__webview.setUrl(QUrl(url))

    def __onload(self, ev):
        self.showNormal()

    def on_create_window(self, func, title:str, url:str, x:int, y:int, width:int, height:int) -> QMainWindow:
        win = func(title, url, x, y, width, height)
        self.__subwindows.append(win)

        return win

    def on_create_dialog(self, func, title:str, url:str, x:int, y:int, width:int, height:int, parent) -> WebDialog:
        dlg = func(title, url, x, y, width, height, self if parent == None else parent)
        self.__subwindows.append(dlg)

        return dlg

def create_window(title:str, url:str, x:int = None, y:int = None, width:int = None, height:int = None):
    global __main_window

    def _create_window(title, url, x, y, width, height):
        return Webwindow(title, url, x, y, width, height)

    if __main_window == None:
        __main_window = Webwindow(title, url, x, y, width, height)
        win = __main_window
    else:
        win = __main_window.create_dialog.emit(_create_window, title, url, x, y, width, height)

    return win

def create_dialog(title:str, url:str, x:int = None, y:int = None, width:int = None, height:int = None, parent:Webwindow = None):
    global __main_window

    def _create_dialog(title, url, x, y, width, height, parent):
        return WebDialog(title, url, x, y, width, height, parent)

    if __main_window == None:
        raise RuntimeError("Please create main Webwindow first!")
    else:
        return __main_window.create_dialog.emit(_create_dialog, title, url, x, y, width, height, parent)

def start(app_icon:str):
    global __qtApp

    __qtApp.setWindowIcon(QIcon(app_icon))
    __qtApp.exec_()



# -*- coding: utf-8 -*-
import os, sys, signal, gc
from collections import OrderedDict
from copy import copy, deepcopy
from flask import Flask, redirect, request, jsonify, __path__ as __flask_path__
import logging, json, json_logging
from datetime import datetime, timedelta
from threading import Thread

from . import __path__, webview
from .logger import Logger
from .interpreter import Interpreter

class Server():
    def __init__(self, app_root:str, enable_logging:bool = True):
        gc.enable()
        self.__logging = enable_logging

        if self.__logging:
            Logger.info("Preparing server...\n")

        self.__app_root = os.path.abspath(app_root)
        self.__session = {
            "server": {
                "mpa": OrderedDict(),
                "upload_time": {}
            },
            "app": {
                "views": OrderedDict(),
                "components": OrderedDict()
            },
            "data": OrderedDict()
        }
        self.__config = {
            "expired_time": 1800
        }

        self.__server = Flask(
            "pyvuejs",
            static_folder = os.path.join(__path__[0], "static")
        )
        self.__server_thread = None
        self.__webview = None
        # self.__server.secret_key = os.urandom(16)

        self.__set_routes()
        self.__get_project_info()

    def __set_routes(self):
        if self.__logging:
            Logger.info("Setting routing points...")

        @self.__server.before_request
        def job_before_request():
            gc.collect()

        @self.__server.after_request
        def job_after_request(response):
            gc.collect()
            return response

        @self.__server.route("/")
        def show_index():
            return redirect("/views/main")

        @self.__server.route("/stop", methods = ["POST"])
        def stop_server():
            if self.__logging:
                Logger.info("Stopping server...")

            sys.exit()

        if self.__logging:
            Logger.info("Setting view/component points...")

        @self.__server.route("/views/<view_name>")
        def show_view(view_name:str):
            if view_name in self.__session["app"]["views"].keys():
                view_id = request.remote_addr
                if not view_id in self.__session["data"].keys():
                    self.__session["data"][view_id] = {}

                if not view_id in self.__session["server"]["mpa"].keys():
                    self.__session["server"]["mpa"][view_id] = {}

                if not view_name in self.__session["server"]["mpa"][view_id].keys():
                    self.__copy_model(view_id, "view", view_name)

                    for model in self.__session["server"]["mpa"][view_id][view_name]["models"].values():
                        if not self.__webview == None:
                            model.webview = self.__webview

                        for var_name, variable in model.sessions.items():
                            self.__session["data"][view_id][var_name] = variable

                        if "load" in model.events.keys():
                            model.call_event("load", self.__session["data"][view_id])

                for model in self.__session["server"]["mpa"][view_id][view_name]["models"].values():
                    if "show" in model.events.keys():
                        model.call_event("show", self.__session["data"][view_id])

                return self.__session["server"]["mpa"][view_id][view_name]["view"].render(
                    view_id,
                    list(self.__session["server"]["mpa"][view_id][view_name]["models"].keys())
                )
            else:
                return ""

        @self.__server.route("/components/<component_name>")
        def show_component(component_name:str):
            if component_name in self.__session["app"]["components"].keys():
                component_id = request.remote_addr
                if not component_id in self.__session["data"].keys():
                    self.__session["data"][component_id] = {}

                if not component_id in self.__session["server"]["mpa"].keys():
                    self.__session["server"]["mpa"][component_id] = OrderedDict()

                if not component_name in self.__session["server"]["mpa"][component_id].keys():
                    self.__copy_model(component_id, "component", component_name)

                    for model in self.__session["server"]["mpa"][component_id][component_name]["models"].values():
                        if not self.__webview == None:
                            model.webview = self.__webview

                        for var_name, variable in model.sessions.items():
                            self.__session["data"][component_id][var_name] = variable

                        if "load" in model.events.keys():
                            model.call_event("load", self.__session["data"][component_id])

                for model in self.__session["server"]["mpa"][component_id][component_name]["models"].values():
                    if "show" in model.events.keys():
                        model.call_event("show", self.__session["data"][component_id])

                return self.__session["server"]["mpa"][component_id][component_name]["view"].render(
                    component_id,
                    list(self.__session["server"]["mpa"][component_id][component_name]["models"].keys())
                )
            else:
                return ""

        if self.__logging:
            Logger.info("Setting function points...")

        @self.__server.route("/functions/init_view", methods = ["POST"])
        def job_init_view():
            res = json.loads(request.data)

            if res["id"] in self.__session["server"]["mpa"].keys():
                view_info = self.__session["server"]["mpa"][res["id"]][res["name"]]
                return jsonify({
                    "state": "success",
                    "session": self.__session["data"][res["id"]],
                    "data": {
                        model_name: model.variables
                        for model_name, model in view_info["models"].items()
                    },
                    "methods": {
                        model_name: list(model.methods.keys())
                        for model_name, model in view_info["models"].items()
                    }
                })

            return jsonify({ "state": "failed" })

        @self.__server.route("/functions/method", methods = ["POST"])
        def job_call_method():
            res = json.loads(request.data)

            if res["id"] in self.__session["server"]["mpa"].keys():
                view_info = self.__session["server"]["mpa"][res["id"]][res["name"]]
                model = view_info["models"][res["model"]]

                model.call_method(res["method"], self.__session["data"][res["id"]])

                return jsonify({ "state": "success" })

            return jsonify({ "state": "failed" })

        @self.__server.route("/functions/upload_variable", methods = ["POST"])
        def job_upload_variable():
            res = json.loads(request.data)

            if res["id"] in self.__session["server"]["mpa"].keys():
                view_info = self.__session["server"]["mpa"][res["id"]][res["name"]]
                for model_name, variables in res["variable"].items():
                    model = view_info["models"][model_name]
                    for var_name, value in variables.items():
                        model.variables[var_name] = value

                return jsonify({ "state": "success" })

            return jsonify({ "state": "failed" })

        @self.__server.route("/functions/download_variable", methods = ["POST"])
        def job_download_variable():
            res = json.loads(request.data)

            if res["id"] in self.__session["server"]["mpa"].keys():
                view_info = self.__session["server"]["mpa"][res["id"]][res["name"]]
                return jsonify({
                    "state": "success",
                    "variable": {
                        model_name: model.variables
                        for model_name, model in view_info["models"].items()
                    }
                })

            return jsonify({ "state": "failed" })

        @self.__server.route("/functions/upload_session", methods = ["POST"])
        def job_upload_session():
            res = json.loads(request.data)

            if res["id"] in self.__session["data"].keys():
                self.__session["data"][res["id"]] = res["session"]
                self.__session["server"]["upload_time"][res["id"]] = datetime.now()

                return jsonify({ "state": "success" })

            return jsonify({ "state": "failed" })

        @self.__server.route("/functions/download_session", methods = ["POST"])
        def job_download_session():
            res = json.loads(request.data)

            for view_id in self.__session["data"].keys():
                if view_id in self.__session["server"]["upload_time"]:
                    last_update_time = self.__session["server"]["upload_time"][view_id]
                    time_delta = timedelta(seconds = self.__config["expired_time"])

                    if last_update_time + time_delta < datetime.now():
                        self.__session["data"].pop(view_id)
                        self.__session["server"]["mpa"].pop(view_id)
                else:
                    self.__session["server"]["upload_time"][view_id] = datetime.now()

            if res["id"] in self.__session["data"].keys():
                return jsonify({
                    "state": "success",
                    "session": self.__session["data"][res["id"]]
                })

            return jsonify({ "state": "failed" })

        if self.__logging:
            Logger.info("finished\n")

    def __get_project_info(self):
        project_info = Interpreter(self.__app_root, self.__server, self.__logging).interprete()
        for view_name, view_info in project_info.items():
            self.__session["app"]["{}s".format(view_info["view"].prefix)][view_name] = view_info

    def __copy_model(self, view_id:str, view_prefix:str, view_name:str):
        try:
            self.__session["server"]["mpa"][view_id][view_name] = deepcopy(self.__session["app"]["{}s".format(view_prefix)][view_name])
        except TypeError as err:
            if err.args[0].startswith("can't pickle"):
                self.__session["server"]["mpa"][view_id][view_name] = copy(self.__session["app"]["{}s".format(view_prefix)][view_name])

    def __start_thread(self):
        self.__server.run(
            host = self.__config["host"],
            port = self.__config["port"],
            debug = False
        )

    def start(self, host:str, port:int, no_wait:bool = False):
        logging.getLogger("werkzeug").setLevel(logging.ERROR)

        self.__config["host"] = "0.0.0.0" if host in ("127.0.0.1", "localhost") else host
        self.__config["port"] = port
        if self.__logging:
            Logger.info("Server started on \"http://{}:{}/\"".format(
                "127.0.0.1" if host == "0.0.0.0" else host, port
            ))
            Logger.info("Please check Devtool to show data transfers")

        self.__server_thread = Thread(target = self.__start_thread)
        self.__server_thread.start()

        if not no_wait:
            try:
                while True:
                    pass
            except KeyboardInterrupt:
                pass

            self.stop()

    def start_standalone(self, host:str, port:int):
        Logger.info("Start server on background...")
        self.start(host, port, True)

        Logger.info("Setting up webview...")
        webview_icon = os.path.join(self.__app_root, "static", "favicon.png")
        if not os.path.exists(webview_icon):
            webview_icon = os.path.join(__path__[0], "static", "favicon.png")

        self.__webview = webview.create_window(
            os.path.basename(self.__app_root),
            "http://127.0.0.1:{}/".format(port),
            width = 950, height = 650
        )

        Logger.info("Webview is loaded")
        webview.start(webview_icon)

        Logger.info("Shutting down background server...")
        self.stop()

    def stop(self):
        import sqlite3
        con = sqlite3.connect(os.path.join(self.__app_root, ".config"))
        cursor = con.cursor()
        cursor.execute("delete from `state`;")
        cursor.close()
        con.close()

        os.kill(os.getpid(), signal.SIGTERM)
